package java.lang.invoke;

import java.util.Arrays;

class MethodHandleNatives
{
    static final boolean COUNT_GWT;
    
    private MethodHandleNatives() {
    }
    
    static native void init(MemberName p0, Object p1);
    
    static native void expand(MemberName p0);
    
    static native MemberName resolve(MemberName p0, Class<?> p1) throws LinkageError, ClassNotFoundException;
    
    static native int getMembers(Class<?> p0, String p1, String p2, int p3, Class<?> p4, int p5, MemberName[] p6);
    
    static native long objectFieldOffset(MemberName p0);
    
    static native long staticFieldOffset(MemberName p0);
    
    static native Object staticFieldBase(MemberName p0);
    
    static native Object getMemberVMInfo(MemberName p0);
    
    static native int getConstant(int p0);
    
    static native void setCallSiteTargetNormal(CallSite p0, MethodHandle p1);
    
    static native void setCallSiteTargetVolatile(CallSite p0, MethodHandle p1);
    
    private static native void registerNatives();
    
    static boolean refKindIsValid(int n) {
        return n > 0 && n < 10;
    }
    
    static boolean refKindIsField(byte b) {
        assert !(!refKindIsValid((int)b));
        return b <= 4;
    }
    
    static boolean refKindIsGetter(byte b) {
        assert !(!refKindIsValid((int)b));
        return b <= 2;
    }
    
    static boolean refKindIsSetter(byte b) {
        return refKindIsField(b) && !refKindIsGetter(b);
    }
    
    static boolean refKindIsMethod(byte b) {
        return !refKindIsField(b) && b != 8;
    }
    
    static boolean refKindIsConstructor(byte b) {
        return b == 8;
    }
    
    static boolean refKindHasReceiver(byte b) {
        assert !(!refKindIsValid((int)b));
        return (b & 0x1) != 0x0;
    }
    
    static boolean refKindIsStatic(byte b) {
        return !refKindHasReceiver(b) && b != 8;
    }
    
    static boolean refKindDoesDispatch(byte b) {
        assert !(!refKindIsValid((int)b));
        return b == 5 || b == 9;
    }
    
    static String refKindName(byte b) {
        assert !(!refKindIsValid((int)b));
        switch (b) {
            case 1: {
                return "getField";
            }
            case 2: {
                return "getStatic";
            }
            case 3: {
                return "putField";
            }
            case 4: {
                return "putStatic";
            }
            case 5: {
                return "invokeVirtual";
            }
            case 6: {
                return "invokeStatic";
            }
            case 7: {
                return "invokeSpecial";
            }
            case 8: {
                return "newInvokeSpecial";
            }
            case 9: {
                return "invokeInterface";
            }
            default: {
                return "REF_???";
            }
        }
    }
    
    private static native int getNamedCon(int p0, Object[] p1);
    
    static boolean verifyConstants() {
        Object[] array = { null };
        int n = 0;
        while (true) {
            array[0] = null;
            int namedCon = getNamedCon(n, array);
            if (array[0] == null) {
                break;
            }
            String str = (String)array[0];
            try {
                int int1 = MethodHandleNatives.Constants.class.getDeclaredField(str).getInt(null);
                if (int1 != namedCon) {
                    String string = str + ": JVM has " + namedCon + " while Java has " + int1;
                    if (str.equals("CONV_OP_LIMIT")) {
                        System.err.println("warning: " + string);
                    }
                    else {
                        throw new InternalError(string);
                    }
                }
            }
            catch (NoSuchFieldException | IllegalAccessException ex) {
                new StringBuilder().append(str).append(": JVM has ").append(namedCon).append(" which Java does not define").toString();
            }
            ++n;
        }
        return true;
    }
    
    static MemberName linkCallSite(Object o, Object o2, Object o3, Object o4, Object o5, Object[] array) {
        MethodHandle methodHandle = (MethodHandle)o2;
        Class<Object> clazz = (Class)o;
        String intern = o3.toString().intern();
        MethodType methodType = (MethodType)o4;
        if (!MethodHandleStatics.TRACE_METHOD_LINKAGE) {
            return linkCallSiteImpl(clazz, methodHandle, intern, methodType, o5, array);
        }
        return linkCallSiteTracing(clazz, methodHandle, intern, methodType, o5, array);
    }
    
    static MemberName linkCallSiteImpl(Class<?> clazz, MethodHandle methodHandle, String s, MethodType methodType, Object o, Object[] array) {
        CallSite site = CallSite.makeSite(methodHandle, s, methodType, o, clazz);
        if ((site instanceof ConstantCallSite)) {
            array[0] = site.dynamicInvoker();
            return Invokers.linkToTargetMethod(methodType);
        }
        array[0] = site;
        return Invokers.linkToCallSiteMethod(methodType);
    }
    
    static MemberName linkCallSiteTracing(Class<?> clazz, MethodHandle methodHandle, String str, MethodType obj, Object obj2, Object[] array) {
        MemberName internalMemberName = methodHandle.internalMemberName();
        if (internalMemberName == null) {
            internalMemberName = methodHandle;
        }
        obj2 = ((obj2 instanceof Object[]) ? Arrays.asList((Object[])(Object[])obj2) : obj2);
        System.out.println("linkCallSite " + clazz.getName() + " " + internalMemberName + " " + str + obj + "/" + obj2);
        try {
            MemberName linkCallSiteImpl = linkCallSiteImpl(clazz, methodHandle, str, obj, obj2, array);
            System.out.println("linkCallSite => " + linkCallSiteImpl + " + " + array[0]);
            return linkCallSiteImpl;
        }
        catch (Throwable obj3) {
            System.out.println("linkCallSite => throw " + obj3);
            throw obj3;
        }
    }
    
    static MethodType findMethodHandleType(Class<?> clazz, Class<?>[] array) {
        return MethodType.makeImpl(clazz, array, true);
    }
    
    static MemberName linkMethod(Class<?> clazz, int n, Class<?> clazz2, String s, Object o, Object[] array) {
        if (!MethodHandleStatics.TRACE_METHOD_LINKAGE) {
            return linkMethodImpl(clazz, n, clazz2, s, o, array);
        }
        return linkMethodTracing(clazz, n, clazz2, s, o, array);
    }
    
    static MemberName linkMethodImpl(Class<?> clazz, int n, Class<?> clazz2, String str, Object obj, Object[] array) {
        try {
            if (clazz2 == MethodHandle.class && n == 5) {
                return Invokers.methodHandleInvokeLinkerMethod(str, fixMethodType(clazz, obj), array);
            }
        }
        catch (Throwable cause) {
            if ((cause instanceof LinkageError)) {
                throw (LinkageError)cause;
            }
            throw new LinkageError(cause.getMessage(), cause);
        }
        throw new LinkageError("no such method " + clazz2.getName() + "." + str + obj);
    }
    
    private static MethodType fixMethodType(Class<?> clazz, Object o) {
        if ((o instanceof MethodType)) {
            return (MethodType)o;
        }
        return MethodType.fromMethodDescriptorString((String)o, clazz.getClassLoader());
    }
    
    static MemberName linkMethodTracing(Class<?> clazz, int i, Class<?> clazz2, String str, Object obj, Object[] array) {
        System.out.println("linkMethod " + clazz2.getName() + "." + str + obj + "/" + Integer.toHexString(i));
        try {
            MemberName linkMethodImpl = linkMethodImpl(clazz, i, clazz2, str, obj, array);
            System.out.println("linkMethod => " + linkMethodImpl + " + " + array[0]);
            return linkMethodImpl;
        }
        catch (Throwable obj2) {
            System.out.println("linkMethod => throw " + obj2);
            throw obj2;
        }
    }
    
    static MethodHandle linkMethodHandleConstant(Class<?> clazz, int n, Class<?> clazz2, String s, Object o) {
        try {
            MethodHandles.Lookup in = MethodHandles.Lookup.IMPL_LOOKUP.in(clazz);
            assert !(!refKindIsValid(n));
            return in.linkMethodHandleConstant((byte)n, clazz2, s, o);
        }
        catch (IllegalAccessException ex) {
            Throwable cause = ex.getCause();
            if ((cause instanceof AbstractMethodError)) {
                throw (AbstractMethodError)cause;
            }
            throw initCauseFrom(new IllegalAccessError(ex.getMessage()), ex);
        }
        catch (NoSuchMethodException ex2) {
            throw initCauseFrom(new NoSuchMethodError(ex2.getMessage()), ex2);
        }
        catch (NoSuchFieldException ex3) {
            throw initCauseFrom(new NoSuchFieldError(ex3.getMessage()), ex3);
        }
        catch (ReflectiveOperationException ex4) {
            throw initCauseFrom(new IncompatibleClassChangeError(), ex4);
        }
    }
    
    private static Error initCauseFrom(Error error, Exception ex) {
        Throwable cause = ex.getCause();
        if (error.getClass().isInstance(cause)) {
            return (Error)cause;
        }
        error.initCause((cause == null) ? ex : cause);
        return error;
    }
    
    static boolean isCallerSensitive(MemberName memberName) {
        return memberName.isInvocable() && (memberName.isCallerSensitive() || canBeCalledVirtual(memberName));
    }
    
    static boolean canBeCalledVirtual(MemberName memberName) {
        assert !(!memberName.isInvocable());
        memberName.getDeclaringClass();
        String name = memberName.getName();
        int n = -1;
        switch (name.hashCode()) {
            case 622280134: {
                if (name.equals("checkMemberAccess")) {
                    n = 0;
                    break;
                }
                break;
            }
            case 1178897522: {
                if (name.equals("getContextClassLoader")) {
                    n = 1;
                    break;
                }
                break;
            }
        }
        switch (n) {
            case 0: {
                return canBeCalledVirtual(memberName, SecurityManager.class);
            }
            case 1: {
                return canBeCalledVirtual(memberName, Thread.class);
            }
            default: {
                return false;
            }
        }
    }
    
    static boolean canBeCalledVirtual(MemberName memberName, Class<?> clazz) {
        Class<Object> declaringClass = memberName.getDeclaringClass();
        return declaringClass == clazz || (!memberName.isStatic() && !memberName.isPrivate() && (clazz.isAssignableFrom(declaringClass) || declaringClass.isInterface()));
    }
    
    static {
        registerNatives();
        COUNT_GWT = (getConstant(4) != 0);
        MethodHandleImpl.initStatics();
        int i = 1;
        while (i < 10) {
            assert refKindHasReceiver((byte)i) == ((1 << i & 0x2AA) != 0x0) : i;
            i = (byte)(i + 1);
        }
        assert !(!verifyConstants());
        return;
    }
    
    static class Constants
    {
        static final int GC_COUNT_GWT = 4;
        static final int GC_LAMBDA_SUPPORT = 5;
        static final int MN_IS_METHOD = 65536;
        static final int MN_IS_CONSTRUCTOR = 131072;
        static final int MN_IS_FIELD = 262144;
        static final int MN_IS_TYPE = 524288;
        static final int MN_CALLER_SENSITIVE = 1048576;
        static final int MN_REFERENCE_KIND_SHIFT = 24;
        static final int MN_REFERENCE_KIND_MASK = 15;
        static final int MN_SEARCH_SUPERCLASSES = 1048576;
        static final int MN_SEARCH_INTERFACES = 2097152;
        static final int T_BOOLEAN = 4;
        static final int T_CHAR = 5;
        static final int T_FLOAT = 6;
        static final int T_DOUBLE = 7;
        static final int T_BYTE = 8;
        static final int T_SHORT = 9;
        static final int T_INT = 10;
        static final int T_LONG = 11;
        static final int T_OBJECT = 12;
        static final int T_VOID = 14;
        static final int T_ILLEGAL = 99;
        static final byte CONSTANT_Utf8 = 1;
        static final byte CONSTANT_Integer = 3;
        static final byte CONSTANT_Float = 4;
        static final byte CONSTANT_Long = 5;
        static final byte CONSTANT_Double = 6;
        static final byte CONSTANT_Class = 7;
        static final byte CONSTANT_String = 8;
        static final byte CONSTANT_Fieldref = 9;
        static final byte CONSTANT_Methodref = 10;
        static final byte CONSTANT_InterfaceMethodref = 11;
        static final byte CONSTANT_NameAndType = 12;
        static final byte CONSTANT_MethodHandle = 15;
        static final byte CONSTANT_MethodType = 16;
        static final byte CONSTANT_InvokeDynamic = 18;
        static final byte CONSTANT_LIMIT = 19;
        static final char ACC_PUBLIC = '\u0001';
        static final char ACC_PRIVATE = '\u0002';
        static final char ACC_PROTECTED = '\u0004';
        static final char ACC_STATIC = '\b';
        static final char ACC_FINAL = '\u0010';
        static final char ACC_SYNCHRONIZED = ' ';
        static final char ACC_VOLATILE = '@';
        static final char ACC_TRANSIENT = '\u0080';
        static final char ACC_NATIVE = 'Ā';
        static final char ACC_INTERFACE = 'Ȁ';
        static final char ACC_ABSTRACT = 'Ѐ';
        static final char ACC_STRICT = 'ࠀ';
        static final char ACC_SYNTHETIC = 'က';
        static final char ACC_ANNOTATION = '\u2000';
        static final char ACC_ENUM = '䀀';
        static final char ACC_SUPER = ' ';
        static final char ACC_BRIDGE = '@';
        static final char ACC_VARARGS = '\u0080';
        static final byte REF_NONE = 0;
        static final byte REF_getField = 1;
        static final byte REF_getStatic = 2;
        static final byte REF_putField = 3;
        static final byte REF_putStatic = 4;
        static final byte REF_invokeVirtual = 5;
        static final byte REF_invokeStatic = 6;
        static final byte REF_invokeSpecial = 7;
        static final byte REF_newInvokeSpecial = 8;
        static final byte REF_invokeInterface = 9;
        static final byte REF_LIMIT = 10;
    }
}
